import errno
import logging
import os
from collections import namedtuple
from math import ceil
from pathlib import Path

from .utils import (
    can_potentially_miss_in_unsplitable,
    get_root_of_unsplitable,
    is_unsplitable,
    parse_torrent,
)

MatchedFile = namedtuple("MatchedFile", ["torrent_file", "searched_files"])
MatchResult = namedtuple("MatchResult", ["root_path", "matched_files", "size"])
MappedFile = namedtuple("MappedFile", ["size", "clients"])
MapResult = namedtuple("MapResult", ["total_size", "seeded_size", "files"])
DynamicMatchResult = namedtuple(
    "DynamicMatchResult", ["success", "missing_size", "matched_files", "touched_files"]
)

logger = logging.getLogger(__name__)

EXACT_MATCH_FACTOR = 0.05


def is_relative_to(path, *other):
    """Return True if the path is relative to another path or False."""
    try:
        path.relative_to(*other)
        return True
    except ValueError:
        return False


class Matcher:
    def __init__(self, rewriter, db):
        self.rewriter = rewriter
        self.db = db

    def _match_filelist_exact(
        self,
        filelist,
        skip_prefix_path=None,
        match_normalized_filename=False,
    ):
        if skip_prefix_path:
            skip_prefix_path = Path(skip_prefix_path.strip(os.sep))
            filelist = [f for f in filelist if is_relative_to(f.path, skip_prefix_path)]
        filelist = sorted(
            filelist,
            key=lambda f: (not can_potentially_miss_in_unsplitable(f.path), f.size),
            reverse=True,
        )

        if not filelist:
            logger.warning(
                f"Empty filelist, bailing - skip_prefix_path:{skip_prefix_path}"
            )
            return None

        handled_root_paths = set()
        match_results = []
        for search_file in filelist[: ceil(len(filelist) * EXACT_MATCH_FACTOR)]:
            path_postfix = search_file.path.parent
            if match_normalized_filename:
                entry_matched_files = self.db.search_file(
                    normalized_filename=search_file.path.name,
                    size=search_file.size,
                    path_postfix=path_postfix,
                )
            else:
                entry_matched_files = self.db.search_file(
                    filename=search_file.path.name,
                    size=search_file.size,
                    path_postfix=path_postfix,
                )
            for entry_matched_file in entry_matched_files:
                if search_file.path:
                    root_path = entry_matched_file.path
                    for _ in search_file.path.parts[1:]:
                        root_path = root_path.parent
                else:
                    root_path = entry_matched_file.path
                if root_path in handled_root_paths:
                    logger.debug(
                        f"Skipping scan of root_path {handled_root_paths} for matches"
                    )
                    continue
                handled_root_paths.add(root_path)
                logger.debug(f"Scanning root_path {root_path} for matches")

                matched_files = [MatchedFile(search_file, [entry_matched_file])]
                matched_file_size = entry_matched_file.size
                for f in filelist:
                    if f == search_file:
                        continue
                    f_path = root_path / f.path
                    f_name = f_path.name
                    f_path = f_path.parent
                    if match_normalized_filename:
                        search_result = self.db.search_file(
                            normalized_filename=f_name, size=f.size, path=f_path
                        )
                    else:
                        search_result = self.db.search_file(
                            filename=f_name, size=f.size, path=f_path
                        )
                    matched_files.append(MatchedFile(f, search_result))
                    if search_result:
                        matched_file_size += f.size
                match_results.append(
                    MatchResult(root_path, matched_files, matched_file_size)
                )

        return match_results

    def _match_filelist_unsplitable(
        self,
        filelist,
        skip_prefix_path=None,
        match_normalized_filename=False,
    ):
        if skip_prefix_path:
            skip_prefix_path = Path(skip_prefix_path.strip(os.sep))
            filelist = [f for f in filelist if is_relative_to(f.path, skip_prefix_path)]
        filelist = sorted(
            filelist,
            key=lambda f: (not can_potentially_miss_in_unsplitable(f.path), f.size),
            reverse=True,
        )

        if not filelist:
            logger.warning(
                f"Empty filelist, bailing - skip_prefix_path:{skip_prefix_path}"
            )
            return None

        handled_root_paths = set()
        match_results = []
        for search_file in filelist[: ceil(len(filelist) * EXACT_MATCH_FACTOR)]:
            relative_path = search_file.path.relative_to(skip_prefix_path)
            if match_normalized_filename:
                entry_matched_files = self.db.search_file(
                    normalized_filename=search_file.path.name,
                    size=search_file.size,
                    path_postfix=relative_path.parent,
                )
            else:
                entry_matched_files = self.db.search_file(
                    filename=search_file.path.name,
                    size=search_file.size,
                    path_postfix=relative_path.parent,
                )

            for entry_matched_file in entry_matched_files:
                root_path = entry_matched_file.path
                for _ in range(len(relative_path.parts) - 1):
                    root_path = root_path.parent
                if root_path in handled_root_paths:
                    logger.debug(
                        f"Skipping scan of root_path {handled_root_paths} for matches"
                    )
                    continue
                handled_root_paths.add(root_path)

                root_path_is_correct_name = root_path.name == skip_prefix_path.name
                logger.debug(
                    f"Scanning root_path {root_path} for matches with root_path_is_correct_name={root_path_is_correct_name}"
                )

                matched_files = [MatchedFile(search_file, [entry_matched_file])]
                matched_file_size = entry_matched_file.size

                bad_path_found = False
                for f in filelist:
                    if f == search_file:
                        continue
                    f_path = root_path / f.path.relative_to(skip_prefix_path)
                    f_name = f_path.name
                    f_path = f_path.parent
                    if match_normalized_filename:
                        search_result = self.db.search_file(
                            normalized_filename=f_name, size=f.size, path=f_path
                        )
                    else:
                        search_result = self.db.search_file(
                            filename=f_name, size=f.size, path=f_path
                        )
                    matched_files.append(MatchedFile(f, search_result))
                    if search_result:
                        matched_file_size += f.size

                    if (
                        not search_result
                        and not root_path_is_correct_name
                        and not can_potentially_miss_in_unsplitable(f.path)
                    ):
                        bad_path_found = True
                        break

                if bad_path_found:
                    logger.debug(f"Bad path found wit root_path={root_path}")
                    continue

                match_results.append(
                    MatchResult(root_path, matched_files, matched_file_size)
                )

        return match_results

    def _match_best_file(
        self,
        torrent,
        torrent_file,
        searched_files,
        hash_probe=False,
        match_hash_size=False,
    ):
        searched_files = sorted(
            searched_files, key=lambda x: x.name == torrent_file.path.name, reverse=True
        )
        for searched_file in searched_files:
            if hash_probe:
                searched_file_path = searched_file.path / searched_file.name
                with searched_file_path.open("rb") as fp:
                    matched_hash_probe = torrent_file.pieces.probe_hash(
                        searched_file.size, fp
                    )
                    if (
                        matched_hash_probe is False
                        or not matched_hash_probe
                        and match_hash_size
                    ):
                        logger.debug(
                            f"File {searched_file_path} matched against {torrent_file.path} failed hash probe, skipping"
                        )
                        continue
            return searched_file
        return None

    def _select_best_candidate(
        self, torrent, candidates, hash_probe=False, match_hash_size=False
    ):
        evaluated_candidates = []
        for match_result in candidates:
            candidate_result = {}
            for matched_file in match_result.matched_files:
                candidate_result[
                    matched_file.torrent_file.path
                ] = self._match_best_file(
                    torrent,
                    matched_file.torrent_file,
                    matched_file.searched_files,
                    hash_probe=hash_probe,
                    match_hash_size=match_hash_size,
                )
            evaluated_candidates.append(candidate_result)
        return sorted(
            evaluated_candidates,
            key=lambda x: sum(y.size for y in x.values() if y is not None),
            reverse=True,
        )[0]

    def match_files_exact(self, torrent):
        torrent = parse_torrent(torrent, utf8_compat_mode=self.db.utf8_compat_mode)
        logger.info(f"Doing exact lookup for {torrent}")
        match_results = self._match_filelist_exact(torrent.filelist)
        usable_match_results = []
        for match_result in match_results:
            if any(
                not matched_file.searched_files
                for matched_file in match_result.matched_files
            ):
                logger.debug("Match with missing files found, skipping")
                continue
            usable_match_results.append(match_result)
        if not usable_match_results:
            logger.info(f"No exact match found for {torrent}")
            return None

        return usable_match_results[0].root_path

    def match_files_dynamic(
        self,
        torrent,
        match_hash_size=False,
        add_limit_size=0,
        add_limit_percent=0,
        hash_probe=False,
    ):
        if match_hash_size:
            hash_probe = True
        torrent = parse_torrent(torrent, utf8_compat_mode=self.db.utf8_compat_mode)

        path_files = {}
        for f in torrent.filelist:
            path_files.setdefault(f.path.parent, []).append(f)

        unsplitable_roots = set()
        for path, files in path_files.items():
            parts = path.parts
            while parts:
                if parts in unsplitable_roots:
                    break
                parts = parts[:-1]
            else:
                if is_unsplitable([f.path for f in files]):
                    unsplitable_root = get_root_of_unsplitable(Path(path))
                    unsplitable_roots.add(unsplitable_root.parts)

        best_possible_size = 0
        candidate_paths = {}
        for unsplitable_root in unsplitable_roots:
            # Unsplitable paths cannot be matched with hash_size, this is
            # because it will often contain lots of file of the same size
            # and there would be too many candidates.
            # match_results = self._match_filelist_exact(
            #     torrent.filelist,
            #     skip_prefix_path=os.path.sep.join(unsplitable_root),
            #     match_normalized_filename=True,
            # )
            match_results = self._match_filelist_unsplitable(
                torrent.filelist,
                skip_prefix_path=os.path.sep.join(unsplitable_root),
                match_normalized_filename=True,
            )
            candidate_paths[unsplitable_root] = sorted(
                match_results, key=lambda x: -x.size
            )
            if candidate_paths[unsplitable_root]:
                best_possible_size += candidate_paths[unsplitable_root][0].size

        candidate_files = {}
        for path, files in path_files.items():
            parts = path.parts
            while parts:
                if parts in unsplitable_roots:
                    break
                parts = parts[:-1]
            else:
                for torrent_file in files:
                    if match_hash_size:
                        searched_files = self.db.search_file(size=torrent_file.size)
                    else:
                        searched_files = self.db.search_file(
                            normalized_filename=torrent_file.path.name,
                            size=torrent_file.size,
                        )
                    candidate_files[torrent_file.path] = (torrent_file, searched_files)
                    if searched_files:
                        best_possible_size += torrent_file.size

        max_missing_size = min(
            add_limit_size, (add_limit_percent * torrent.size) // 100
        )
        current_missing_size = torrent.size - best_possible_size
        if current_missing_size > max_missing_size:
            logger.info(
                f"Torrent missing too much data, size:{torrent.size}, found data size:{best_possible_size}"
            )
            return DynamicMatchResult(False, current_missing_size, None, None)

        result_mapping = {}
        for path, candidates in candidate_paths.items():
            result_mapping.update(
                self._select_best_candidate(
                    torrent,
                    candidates,
                    hash_probe=hash_probe,
                    match_hash_size=match_hash_size,
                )
            )

        for path, (torrent_file, searched_files) in candidate_files.items():
            result_mapping[torrent_file.path] = self._match_best_file(
                torrent,
                torrent_file,
                searched_files,
                hash_probe=hash_probe,
                match_hash_size=match_hash_size,
            )

        missing_pieces = set()
        found_pieces = set()
        found_file_piece_mapping = {}
        current_missing_size = 0
        for torrent_file in torrent.filelist:
            piece_calculation = torrent_file.pieces.calculate_offsets(torrent_file.size)
            if result_mapping[torrent_file.path]:
                found_pieces.add(piece_calculation.start_piece)
                found_pieces.add(piece_calculation.end_piece)
                found_file_piece_mapping.setdefault(
                    piece_calculation.start_piece, []
                ).append(torrent_file.path)
                found_file_piece_mapping.setdefault(
                    piece_calculation.end_piece, []
                ).append(torrent_file.path)
            else:
                missing_pieces.add(piece_calculation.start_piece)
                missing_pieces.add(piece_calculation.end_piece)
                current_missing_size += torrent_file.size

        if current_missing_size > max_missing_size:
            logger.info(
                f"Torrent missing too much data after matching files, size:{torrent.size}, found data size:{best_possible_size} missing size:{current_missing_size} max missing size:{max_missing_size}"
            )
            return DynamicMatchResult(False, current_missing_size, None, None)

        touched_files = set()
        for piece in missing_pieces & found_pieces:
            touched_files |= set(found_file_piece_mapping[piece])

        return DynamicMatchResult(
            True,
            current_missing_size,
            {
                path: (searched_file and searched_file.to_full_path())
                for (path, searched_file) in result_mapping.items()
            },
            list(touched_files),
        )

    def map_path_to_clients(self, path):
        """
        Map a path and all its files to clients.
        """
        scanned_folders = set()
        total = {"size": 0}
        real_files_seen = set()
        real_files_mapping = {}
        path_seeded = {}
        path_check_queue = []

        def flush_check_queue():
            logger.debug("Flushing queue")
            for p in path_check_queue:
                resolved_p = p.resolve()
                size = p.stat().st_size
                if resolved_p not in real_files_seen:
                    total["size"] += size
                    real_files_seen.add(resolved_p)

                real_files_mapping[p] = resolved_p
                path_seeded[p] = MappedFile(size=size, clients=[])

            for seeded_file in self.db.get_seeded_paths(path_check_queue):
                path_seeded[seeded_file.path].clients.append(
                    (seeded_file.client, seeded_file.infohash)
                )

            path_check_queue.clear()

        def looper(path, initial_path=False):
            if path in scanned_folders:
                return
            logger.debug(f"Scanning path {path!s}")
            scanned_folders.add(path)

            for rewritten_path in self.rewriter.rewrite_path(
                path, prefix_match=initial_path
            ):
                if rewritten_path.is_file():
                    path_check_queue.append(rewritten_path)
                    continue
                try:
                    for p in rewritten_path.iterdir():
                        if p.is_dir():
                            looper(p)
                        elif p.is_file():
                            path_check_queue.append(p)
                except OSError as e:
                    if e.errno != errno.ELOOP:
                        raise e

            if len(path_check_queue) > 1000:
                flush_check_queue()

        looper(path)
        flush_check_queue()
        seeded_size = 0
        already_counted_paths = set()
        for p, mapped_file in path_seeded.items():
            if not mapped_file.clients:
                continue

            resolved_p = real_files_mapping[p]
            if resolved_p in already_counted_paths:
                continue

            already_counted_paths.add(resolved_p)
            seeded_size += mapped_file.size

        return MapResult(
            total_size=total["size"], seeded_size=seeded_size, files=path_seeded
        )



import logging
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from fnmatch import fnmatch
from pathlib import Path
from queue import Empty, SimpleQueue

from .db import InsertTorrentFile
from .utils import get_root_of_unsplitable, is_unsplitable

logger = logging.getLogger(__name__)

INSERT_QUEUE_MAX_SIZE = 1000

SCAN_PATH_QUEUE_TIMEOUT_SECONDS = 10


class IndexAction(Enum):
    ADD = 1
    MARK_UNSPLITABLE = 2
    FINISHED = 3


class Indexer:
    def __init__(self, db, ignore_file_patterns=None, ignore_directory_patterns=None):
        self.db = db
        self.ignore_file_patterns = ignore_file_patterns or []
        self.ignore_directory_patterns = ignore_directory_patterns or []

    def scan_paths(self, paths, full_scan=True):
        if full_scan:
            self.db.truncate_files()
        paths = [Path(p) for p in paths]
        queue = SimpleQueue()
        futures = {}

        with ThreadPoolExecutor(max_workers=len(paths) + 1) as executor:
            for path in paths:
                logger.info(f"Indexing path {path}")
                futures[str(path)] = executor.submit(
                    self._scan_path_thread, path, queue, root_thread=True
                )

            while len(futures):
                action, args = None, ()
                try:
                    action, args = queue.get(timeout=SCAN_PATH_QUEUE_TIMEOUT_SECONDS)
                except Empty:
                    logger.debug(
                        "No action received from queue in %d seconds, checking threads",
                        SCAN_PATH_QUEUE_TIMEOUT_SECONDS,
                    )
                    for path in list(futures):
                        future = futures[path]
                        if future.done():
                            if future.exception() is not None:
                                logger.error(
                                    f"Thread for path {path} encountered an exception: {future.exception()}"
                                )
                            del futures[path]
                if action == IndexAction.ADD:
                    self.db.insert_file_path(*args)
                elif action == IndexAction.MARK_UNSPLITABLE:
                    self.db.mark_unsplitable_root(*args)
                elif action == IndexAction.FINISHED:
                    del futures[args]

        self.db.commit()

    def _match_ignore_pattern(self, ignore_patterns, p, ignore_case=False):
        name = p.name
        if ignore_case:
            name = name.lower()
        for ignore_pattern in ignore_patterns:
            if ignore_case:
                if fnmatch(name, ignore_pattern.lower()):
                    return True
            else:
                if fnmatch(name, ignore_pattern):
                    return True
        return False

    def _scan_path_thread(self, path, queue, root_thread=False):
        files = []
        try:
            for p in path.iterdir():
                if p.is_dir():
                    if self.ignore_directory_patterns and self._match_ignore_pattern(
                        self.ignore_directory_patterns, p, ignore_case=True
                    ):
                        continue
                    self._scan_path_thread(p, queue)
                elif p.is_file():
                    if self.ignore_file_patterns and self._match_ignore_pattern(
                        self.ignore_file_patterns, p
                    ):
                        continue
                    files.append(p)
                    size = p.stat().st_size
                    queue.put((IndexAction.ADD, (p, size)))

            # TODO: probably not utf-8 problems resilient
            if is_unsplitable(files):  # TODO: prevent duplicate work (?)
                unsplitable_root = get_root_of_unsplitable(path)
                queue.put((IndexAction.MARK_UNSPLITABLE, (unsplitable_root,)))
        except OSError as e:
            logger.error(f"Failed to scan {path}: {e}")

        if root_thread:
            queue.put((IndexAction.FINISHED, (str(path))))

    def scan_clients(self, clients, full_scan=False, fast_scan=False):
        for name, client in clients.items():
            if full_scan:
                self.db.truncate_torrent_files(name)
            self._scan_client(name, client, not full_scan and fast_scan)
        self.db.commit()

    def _scan_client(self, client_name, client, fast_scan):
        torrents = client.list()
        insert_queue = []
        for torrent in torrents:
            _, current_download_path = self.db.get_torrent_file_info(
                client_name, torrent.infohash
            )
            if fast_scan and current_download_path is not None:
                logger.debug(
                    f"torrent:{torrent!r} client:{client!r} Skip indexing because it is already there and fast-scan is enabled"
                )
                continue

            download_path = client.get_download_path(torrent.infohash)
            if str(download_path) == current_download_path:
                logger.debug(
                    f"torrent:{torrent!r} client:{client!r} Skip indexing because download path not changed"
                )
                continue

            files = client.get_files(torrent.infohash)
            if not files:
                logger.debug("No files found, not loaded")
            paths = []
            for f in files:
                f_path = download_path / f.path
                paths.append((str(f_path), f.size))
                f_path_resolved = f_path.resolve()
                if f_path_resolved != f_path:
                    paths.append((str(f_path_resolved), f.size))
            insert_queue.append(
                InsertTorrentFile(torrent.infohash, torrent.name, download_path, paths)
            )
            if len(insert_queue) > INSERT_QUEUE_MAX_SIZE:
                self.db.insert_torrent_files_paths(client_name, insert_queue)
                insert_queue = []
        if insert_queue:
            self.db.insert_torrent_files_paths(client_name, insert_queue)

        self.db.remove_non_existing_infohashes(
            client_name, [torrent.infohash for torrent in torrents]
        )



import hashlib
import json
import logging
import shutil
import time
from pathlib import Path

from .utils import chown, create_link

logger = logging.getLogger(__name__)

RW_CACHE_DATA_PATH = "data"
CW_CACHE_CONF_NAME = "autotorrent.json"


class ReadWriteFileCache:
    def __init__(self, path, ttl, chown_str=None):
        self.path = Path(path)
        self.ttl = ttl
        self.chown_str = chown_str

    def cleanup_cache(self):
        removed_paths = []
        for path in self.path.iterdir():
            if time.time() - path.stat().st_mtime > self.ttl:
                logger.debug(f"Path {path} is older than ttl and should be deleted")
                conf_path = path / CW_CACHE_CONF_NAME
                conf = json.loads(conf_path.read_text())
                source_path = Path(conf["source_path"])
                for target_path in conf["target_paths"]:
                    link_type = target_path["link_type"]
                    target_path = Path(target_path["path"])
                    if not target_path.exists():
                        logger.warning(f"Target path {target_path!s} does not exist")
                        continue
                    logger.debug(f"Rewriting {target_path!s} to {source_path!s}")
                    target_path.unlink()
                    create_link(source_path, target_path, link_type)
                removed_paths.append(path)
                shutil.rmtree(path)
        return removed_paths

    def cache_file(self, path, target_path, link_type):
        full_folder_name = "__".join(path.parts[1:])
        folder_name = f"{full_folder_name[:25]}__{full_folder_name[-50:]}__{hashlib.sha1(str(path).encode()).hexdigest()}"
        folder_path = self.path / folder_name
        folder_data_path = folder_path / RW_CACHE_DATA_PATH
        folder_data_file = folder_data_path / path.name
        conf_path = folder_path / CW_CACHE_CONF_NAME
        if not folder_path.exists():
            logger.info(
                f"Seems like folder {folder_path!s} does not exist, copying file from {path!s}"
            )
            folder_path.mkdir()
            folder_data_path.mkdir()
            shutil.copyfile(path, folder_data_file)
            if self.chown_str is not None:
                chown(self.chown_str, folder_data_file)
            conf_path.write_text(
                json.dumps(
                    {
                        "source_path": str(path),
                        "target_paths": [],
                    }
                )
            )

        folder_path.touch()
        conf = json.loads(conf_path.read_text())
        conf["target_paths"].append(
            {
                "path": str(target_path),
                "link_type": link_type,
            }
        )
        conf_path.write_text(json.dumps(conf))
        return folder_data_file



import errno
import hashlib
import json
import logging
import os
import platform
import re
import shutil
from collections import namedtuple
from fnmatch import fnmatch
from pathlib import Path, PurePath

import chardet
import click
from libtc import TorrentProblems

from .exceptions import FailedToCreateLinkException, FailedToParseTorrentException

logger = logging.getLogger(__name__)

UNSPLITABLE_FILE_EXTENSIONS = [
    set([".rar", ".sfv"]),
    set([".rar", ".r00"]),
    set([".mp3", ".sfv"]),
    set([".vob", ".ifo"]),
]
UNSPLITABLE_FILE_MISSABLE = [  # If foldername does not match, these files can be missed
    "*.nfo",
    "*.sfv",
    "*.diz",
    "*.txt",
]

# # Not allowed anywhere in the names
# INVALID_CHARACTERS_NIX = ["/"]
# INVALID_CHARACTERS_WINDOWS = ["/", "<", ">", ":", '"', "\\", "|", "?", "*"]

# # Not allowed as a name or file basename
# INVALID_BASE_NAMES_NIX = []
# INVALID_BASE_NAMES_WINDOWS = [
#     "CON",
#     "PRN",
#     "AUX",
#     "NUL",
#     "COM1",
#     "COM2",
#     "COM3",
#     "COM4",
#     "COM5",
#     "COM6",
#     "COM7",
#     "COM8",
#     "COM9",
#     "LPT1",
#     "LPT2",
#     "LPT3",
#     "LPT4",
#     "LPT5",
#     "LPT6",
#     "LPT7",
#     "LPT8",
#     "LPT9",
# ]

# if os.name == "nt":
#     INVALID_CHARACTERS = INVALID_CHARACTERS_WINDOWS
#     INVALID_BASE_NAMES = INVALID_BASE_NAMES_WINDOWS
# else:
#     INVALID_CHARACTERS = INVALID_CHARACTERS_NIX
#     INVALID_BASE_NAMES = INVALID_BASE_NAMES_NIX

PIECE_SIZE = 20
HASHER_READ_BLOCK_SIZE = 2**18

AUTOTORRENT_CONF_NAME = "autotorrent.json"
STORE_DATA_PATH = "data"


def decode_str(s, try_fix=False):
    orig_s = s
    if not isinstance(s, str) and not isinstance(s, bytes):
        s = str(s)

    if isinstance(s, str):
        try:
            s = s.encode()
        except UnicodeEncodeError:
            if not try_fix:
                return None

    if isinstance(s, bytes):
        try:
            return s.decode()
        except UnicodeDecodeError:
            if not try_fix:
                return None
    else:
        try:
            s = bytes(orig_s)
        except TypeError:
            if not try_fix:
                return None

    encoding = chardet.detect(s)
    if encoding["encoding"]:
        try:
            return s.decode(encoding["encoding"])
        except UnicodeDecodeError:
            pass

    try:
        return os.fsdecode(s)
    except UnicodeDecodeError:
        return s.decode(errors="replace")


def normalize_filename(filename):
    filename = filename.strip(" ")
    name, ext = os.path.splitext(filename)
    name = re.sub(r"[ _.-]+", " ", name)
    return f"{name}{ext}".lower()


def is_unsplitable(files):
    """
    Checks if a list of files can be considered unsplitable, e.g. VOB/IFO or scene release.
    This means the files can only be used in this combination.
    """
    extensions = set(f.suffix.lower() for f in files)
    for exts in UNSPLITABLE_FILE_EXTENSIONS:
        if len(extensions & exts) == len(exts):
            return True

    for f in files:
        if f.name.lower() == "movieobject.bdmv":
            return True

    return False


def can_potentially_miss_in_unsplitable(filepath):
    """
    Checks if a file can be potentially missed in an unsplitable release
    while the release is still usable.
    """
    return re.match(
        r"^((samples?)|(proofs?)|((vob)?sub(title)?s?))$",
        filepath.parent.name,
        re.IGNORECASE,
    ) or any(fnmatch(filepath.name, pattern) for pattern in UNSPLITABLE_FILE_MISSABLE)


def get_root_of_unsplitable(path):
    """
    Scans a path for the actual scene release name, e.g. skipping cd1 folders.

    Returns None if no scene folder could be found
    """
    while path:
        name = path.name
        is_scene_path = re.match(
            r"^((cd[1-9])|(samples?)|(proofs?)|((vob)?sub(title)?s?))$",
            name,
            re.IGNORECASE,
        )
        is_disk_path = re.match(r"^((bdmv)|(disc\d*)|(video_ts))$", name, re.IGNORECASE)
        if (
            not is_disk_path
            and name.lower() == "backup"
            and path.parent.name.lower() == "bdmv"
        ):
            is_disk_path = True

        if not is_scene_path and not is_disk_path:
            return path

        path = path.parent


class PathRewriter:
    def __init__(self, path_mappings):
        self.paths = {}
        self.path_groups = {}
        self.handle_path_mappings(path_mappings)

    def handle_path_mappings(self, path_mappings):
        for i, path_mapping in enumerate(path_mappings):
            for path in path_mapping:
                path = Path(path)
                path_tuple = self._tuplify(path)
                self.paths[path_tuple] = i
                self.path_groups.setdefault(i, []).append(path)

    def rewrite_path(self, path, prefix_match=False):
        postfix_path = None
        orig_path = path
        while path:
            if postfix_path and not prefix_match:  # nothing matched
                break

            path_tuple = self._tuplify(path)
            path_group_id = self.paths.get(path_tuple)
            if path_group_id is not None:
                if postfix_path is None:
                    return list(self.path_groups[path_group_id])
                else:
                    return [p / postfix_path for p in self.path_groups[path_group_id]]

            if postfix_path is None:
                postfix_path = Path(path.name)
            else:
                postfix_path = Path(path.name) / postfix_path

            path = path.parent

        return [orig_path]

    def _tuplify(self, path):
        p = []
        while path.name:
            p.append(path.name)
            path = path.parent
        return tuple(p[::-1])


def validate_path(path):
    return True


def cleanup_torrent_path_segment(path_segment):  # TODO: more here?
    if not path_segment:
        return path_segment
    return path_segment.strip("/")


PieceCalculation = namedtuple(
    "PieceCalculation",
    [
        "start_piece",
        "start_offset",
        "first_complete_piece",
        "end_piece",
        "end_offset",
        "last_complete_piece",
        "pieces",
        "complete_pieces",
    ],
)


class Pieces:
    def __init__(self, piece_length, pieces, start_size=0):
        self.piece_length = piece_length
        if not isinstance(pieces, list):
            pieces = [
                pieces[i : i + PIECE_SIZE] for i in range(0, len(pieces), PIECE_SIZE)
            ]
        self.pieces = pieces
        self.start_size = start_size

    def __getitem__(self, key):
        if not isinstance(key, slice):
            raise TypeError("Must be a slice")

        if not isinstance(key.start, int):
            raise TypeError("The start must be an integer")

        if key.stop is not None:
            raise TypeError("The stop must be None")

        if key.step is not None:
            raise TypeError("The step must be None")

        return self.__class__(
            self.piece_length, self.pieces, self.start_size + key.start
        )

    def hash_piece(self, f):
        """Hashes a full piece from a single file, returns the hash-digest"""
        missing_size = self.piece_length
        hasher = hashlib.sha1()
        logger.debug(f"Trying to read {missing_size} bytes")

        while missing_size:
            d = f.read(min(16384, missing_size))
            if not d:
                logger.warning(
                    f"We expected to be able to read more data with missing size {missing_size}, bailing"
                )
                return None
            missing_size -= len(d)
            hasher.update(d)

        return hasher.digest()

    def calculate_offsets(self, size, is_last_file=False):
        start_piece, start_offset = divmod(self.start_size, self.piece_length)
        first_complete_piece = start_piece
        if start_offset:
            first_complete_piece += 1
            start_offset = self.piece_length - start_offset

        end_size = self.start_size + size
        end_piece, end_offset = divmod(end_size, self.piece_length)
        last_complete_piece = end_piece
        if end_offset and not is_last_file:
            last_complete_piece -= 1

        piece_calculation = PieceCalculation(
            start_piece,
            start_offset,
            first_complete_piece,
            end_piece,
            end_offset,
            last_complete_piece,
            self.pieces[start_piece : end_piece + 1],
            self.pieces[first_complete_piece : last_complete_piece + 1],
        )
        logger.debug(
            f"Piece calculation start_piece: {piece_calculation.start_piece} "
            f"start_offset: {piece_calculation.start_offset} "
            f"first_complete_piece: {piece_calculation.first_complete_piece} "
            f"end_piece: {piece_calculation.end_piece} "
            f"end_offset: {piece_calculation.end_offset} "
            f"last_complete_piece: {piece_calculation.last_complete_piece}"
        )
        return piece_calculation

    def probe_hash(self, size, fp):
        """
        Test a few pieces against the file if possible.

        Returns True if passed, False if failed, None if not possible
        """
        piece_calculation = self.calculate_offsets(size)
        if not piece_calculation.complete_pieces:
            return None

        pieces_to_verify = set([0])
        if len(piece_calculation.complete_pieces) > 1:
            pieces_to_verify.add(len(piece_calculation.complete_pieces) - 1)

        for piece in pieces_to_verify:
            fp.seek(piece_calculation.start_offset + piece * self.piece_length)
            if self.hash_piece(fp) != piece_calculation.complete_pieces[piece]:
                return False

        return True


class Torrent(
    namedtuple(
        "Torrent",
        ["name", "size", "piece_length", "filelist", "filelist_mapped", "trackers"],
    )
):
    def is_problematic(self):
        # TODO: check if the torrent can cause problems with some clients
        return False

    def verify_hash(self, fnmatches, file_mapping):
        """Returns a torrent_file mapping of failed and successful matched files"""
        # loop files, build list of pieces to verify
        pieces_to_verify = set()
        missing_pieces = set()
        for torrent_file in self.filelist:
            piece_calculation = torrent_file.pieces.calculate_offsets(
                torrent_file.size, is_last_file=torrent_file.is_last_file
            )
            torrent_file_pieces = set(
                range(piece_calculation.start_piece, piece_calculation.end_piece + 1)
            )
            for pattern in fnmatches:
                if fnmatch(torrent_file.path.name, pattern):
                    pieces_to_verify |= torrent_file_pieces
                    break
            else:
                if not file_mapping[torrent_file.path]:
                    missing_pieces |= torrent_file_pieces

        piece_status = {}
        file_piece_mapping = {}
        file_has_inner_pieces = {}
        hasher, hasher_piece, data_left, fp, skip_to_piece = (
            None,
            None,
            None,
            None,
            None,
        )
        for torrent_file in self.filelist:
            piece_calculation = torrent_file.pieces.calculate_offsets(
                torrent_file.size, is_last_file=torrent_file.is_last_file
            )
            file_has_inner_pieces[torrent_file] = (
                piece_calculation.first_complete_piece
                <= piece_calculation.last_complete_piece
            )
            full_path = file_mapping[torrent_file.path]
            if not full_path:
                piece_status[piece_calculation.start_piece] = None
                piece_status[piece_calculation.end_piece] = None
                skip_to_piece = piece_calculation.end_piece + 1
                continue

            for piece_index, piece in enumerate(
                piece_calculation.pieces, piece_calculation.start_piece
            ):
                file_piece_mapping.setdefault(piece_index, []).append(torrent_file)
                if skip_to_piece is not None and skip_to_piece > piece_index:
                    continue

                if piece_index not in pieces_to_verify:
                    continue

                if piece_index in piece_status:
                    continue

                if piece_index > piece_calculation.start_piece:
                    expected_tell = piece_calculation.start_offset + (
                        (piece_index - piece_calculation.first_complete_piece)
                        * self.piece_length
                    )
                else:
                    expected_tell = 0

                if not fp:
                    fp = full_path.open("rb")
                    if expected_tell:
                        fp.seek(expected_tell)

                if hasher_piece != piece_index:
                    hasher = hashlib.new("sha1", usedforsecurity=False)
                    hasher_piece = piece_index
                    data_left = min(
                        self.size - (piece_index * self.piece_length), self.piece_length
                    )
                    if fp.tell() != expected_tell:
                        fp.seek(expected_tell)

                while data_left > 0:
                    data = fp.read(min(HASHER_READ_BLOCK_SIZE, data_left))
                    hasher.update(data)
                    data_left -= len(data)
                    if not data:
                        break

                if data_left == 0:
                    piece_status[hasher_piece] = hasher.digest() == piece
                    if not piece_status[hasher_piece]:
                        skip_to_piece = piece_calculation.end_piece

            if fp:
                fp.close()
                fp = None

        file_status_mapping = {}
        for torrent_file in self.filelist:
            for pattern in fnmatches:
                if fnmatch(torrent_file.path.name, pattern):
                    piece_calculation = torrent_file.pieces.calculate_offsets(
                        torrent_file.size, is_last_file=torrent_file.is_last_file
                    )

                    inner_piece_status = [
                        piece_status.get(p)
                        for p in range(
                            piece_calculation.first_complete_piece,
                            piece_calculation.last_complete_piece + 1,
                        )
                    ]
                    edge_piece_status = []
                    if (
                        piece_calculation.start_piece
                        != piece_calculation.first_complete_piece
                    ):
                        edge_piece_status.append(
                            piece_status.get(piece_calculation.start_piece)
                        )
                        # check other files in same piece
                    if (
                        piece_calculation.end_piece
                        != piece_calculation.last_complete_piece
                    ):
                        edge_piece_status.append(
                            piece_status.get(piece_calculation.end_piece)
                        )

                    if (
                        inner_piece_status
                        and all(inner_piece_status)
                        and all([p != False for p in edge_piece_status])
                    ):
                        file_status_mapping[torrent_file] = "hash-success"
                    elif not inner_piece_status and all(edge_piece_status):
                        file_status_mapping[torrent_file] = "hash-success"
                    elif (
                        inner_piece_status
                        and all(inner_piece_status)
                        and all(
                            [
                                file_has_inner_pieces[tf]
                                for tf in file_piece_mapping[
                                    piece_calculation.start_piece
                                ]
                            ]
                        )
                        and all(
                            [
                                file_has_inner_pieces[tf]
                                for tf in file_piece_mapping[
                                    piece_calculation.end_piece
                                ]
                            ]
                        )
                    ):
                        file_status_mapping[torrent_file] = "hash-success"
                    else:
                        file_status_mapping[torrent_file] = "hash-failed"

                    break

        file_touch_status_mapping = {}
        for torrent_file in self.filelist:
            # if hash-failed or any pieces are failed, then it is touch-failed
            # if any of the files in any of the pieces are missing, then it is touched-success
            piece_calculation = torrent_file.pieces.calculate_offsets(
                torrent_file.size, is_last_file=torrent_file.is_last_file
            )
            file_piece_results = {
                piece_status[p]
                for p in range(
                    piece_calculation.start_piece, piece_calculation.end_piece + 1
                )
                if p in piece_status
            }
            if (
                file_status_mapping.get(torrent_file) == "hash-failed"
                or False in file_piece_results
            ):
                file_touch_status_mapping[torrent_file] = "touch-failed"
            elif None in file_piece_results:
                file_touch_status_mapping[torrent_file] = "touch-success"

        return file_status_mapping, file_touch_status_mapping

    def has_file_patterns(self, patterns):
        for torrent_file in self.filelist:
            for pattern in patterns:
                if fnmatch(torrent_file.path.name, pattern):
                    return True
        return False


TorrentFile = namedtuple(
    "TorrentFile", ["path", "size", "pieces", "is_last_file"], defaults=(False,)
)


def parse_torrent(
    torrent, utf8_compat_mode=False
):  # TODO: validate path, add support for transmission rewrite?
    if b"info" not in torrent:
        raise FailedToParseTorrentException("Info dict not found")
    info = torrent[b"info"]
    name = cleanup_torrent_path_segment(
        decode_str(info[b"name"], try_fix=utf8_compat_mode)
    )
    if name is None:
        raise FailedToParseTorrentException("Unable to parse name of torrent")

    pieces = Pieces(info[b"piece length"], info[b"pieces"])
    length = 0
    filelist = []
    if b"files" in info:
        last_i = len(info[b"files"]) - 1
        for i, f in enumerate(info[b"files"]):
            path = [
                cleanup_torrent_path_segment(decode_str(p, try_fix=utf8_compat_mode))
                for p in f[b"path"]
                if p
            ]
            if any(p is None for p in path):
                raise FailedToParseTorrentException(
                    "Broken path elements found in torrent, try utf-8 compat mode"
                )
            if not path:
                raise FailedToParseTorrentException("Empty path")
            if any(p is None for p in path):
                raise FailedToParseTorrentException(
                    "Invalid encoding in torrent file path"
                )
            if any(not validate_path(p) for p in path):
                raise FailedToParseTorrentException(
                    "Illegal entry in torrent file path"
                )

            path = os.path.sep.join([name] + path)
            filelist.append(
                TorrentFile(
                    PurePath(path),
                    f[b"length"],
                    pieces[length:],
                    is_last_file=(i == last_i),
                )
            )
            length += f[b"length"]
    else:
        filelist.append(
            TorrentFile(PurePath(name), info[b"length"], pieces, is_last_file=True)
        )
        length += info[b"length"]

    filelist_mapped = {f.path: f for f in filelist}

    trackers = [torrent.get(b"announce", b"").decode()]
    for tracker_group in torrent.get(b"announce-list", []):
        if not isinstance(tracker_group, list):
            tracker_group = [tracker_group]
        for tracker in tracker_group:
            tracker = tracker.decode()
            if tracker not in trackers:
                trackers.append(trackers)
    trackers = [t for t in trackers if t]
    return Torrent(
        name, length, info[b"piece length"], filelist, filelist_mapped, trackers
    )


CreateLinkResult = namedtuple(
    "CreateLinkResult", ["path", "torrent_path", "config_path", "data_path"]
)


def _parse_chown(chown):
    chown = chown.split(":")
    chown_user = chown[0] or None
    if len(chown) > 1:
        chown_group = chown[1] or None
    else:
        chown_group = None

    try:
        uid = int(chown_user)
    except ValueError:
        uid = shutil._get_uid(chown_user)
        if uid is None:
            uid = -1

    try:
        gid = int(chown_group)
    except ValueError:
        gid = shutil._get_gid(chown_group)
        if gid is None:
            gid = -1
    return uid, gid


def chown(chown_str, path):
    chown_user, chown_group = _parse_chown(chown_str)
    if chown_user == -1 and chown_group == -1:
        return

    logger.debug(f"Got permissions chown_user={chown_user} chown_group={chown_group}")

    def walk(path):
        yield path

        if path.is_dir():
            for p in Path(path).iterdir():
                if p.is_dir() and not p.is_symlink():
                    yield from walk(p)
                else:
                    yield p

    for p in walk(path):
        logger.debug(f"Changing permission for {p}")
        os.chown(p, chown_user, chown_group, follow_symlinks=False)


def create_link_path(
    store_path_template,
    file_mapping,
    client_name,
    torrent_file_path,
    additional_kwargs,
    link_type,
    rw_cache=None,
    chown_str=None,
    dry_run=False,
    skip_store_metadata=False,
):
    kwargs = {
        "client": client_name,
        "torrent_name": torrent_file_path.stem,
    }
    kwargs.update(additional_kwargs)
    store_path = Path(store_path_template.format(**kwargs))

    if dry_run:
        if store_path.exists():
            raise FailedToCreateLinkException(f"Path {store_path} already exist")
        else:
            return None

    try:
        store_path.mkdir(parents=True)
    except FileExistsError:
        raise FailedToCreateLinkException(f"Path {store_path} already exist")

    if skip_store_metadata:
        data_store_path = store_path
        torrent_store_path = None
        autotorrent_store_path = None
    else:
        data_store_path = store_path / STORE_DATA_PATH
        data_store_path.mkdir()

        torrent_store_path = store_path / torrent_file_path.name
        shutil.copy(torrent_file_path, torrent_store_path)

        autotorrent_store_path = store_path / AUTOTORRENT_CONF_NAME
        autotorrent_store_path.write_text(json.dumps({}))

    for torrent_path, (action, actual_path) in file_mapping.items():
        link_path = data_store_path / torrent_path
        link_path.parent.mkdir(exist_ok=True, parents=True)
        if action == "link" or action == "cache_link":
            if rw_cache and action == "cache_link":
                actual_path = rw_cache.cache_file(actual_path, link_path, link_type)

            create_link(actual_path, link_path, link_type)
        elif action == "copy":
            shutil.copyfile(actual_path, link_path)

    if chown_str:
        chown(chown_str, data_store_path)

    return CreateLinkResult(
        store_path, torrent_store_path, autotorrent_store_path, data_store_path
    )


def create_link(actual_path, link_path, link_type):
    if link_type == "soft":
        link_path.symlink_to(actual_path)
    elif link_type == "hard":
        os.link(actual_path, link_path)
    elif link_type == "reflink":
        reflink(str(link_path), str(actual_path))


def reflink(path, destination):
    """
    Perform a reflink (if supported, currently only xfs, apfs, btrfs is)
    This code is modified from dvc (https://github.com/iterative/dvc/blob/f4bec650eddc8874b3f7ab2f8b34bc5dfe60fd49/dvc/system.py#L105).
    These libraries are available under the Apache 2.0 license, which can be obtained from http://www.apache.org/licenses/LICENSE-2.0.
    """
    system = platform.system()
    logger.debug(f"platform is {system}")
    try:
        if system == "Windows":
            ret = _reflink_windows(path, destination)
        elif system == "Darwin":
            ret = _reflink_darwin(path, destination)
        elif system == "Linux":
            ret = _reflink_linux(path, destination)
        else:
            ret = -1
    except IOError:
        ret = -1

    if ret != 0:
        raise Exception("reflink is not supported")


def _reflink_linux(self, path, destination):
    """
    Linux only reflink via syscall FICLONE on supported filesystems
    """
    import fcntl
    import os

    FICLONE = 0x40049409

    try:
        ret = 255
        with open(path, "r") as s, open(destination, "w+") as d:
            ret = fcntl.ioctl(d.fileno(), FICLONE, s.fileno())
    finally:
        if ret != 0:
            os.unlink(destination)

    return ret


def _reflink_windows(self, path, destination):
    return -1


def _reflink_darwin(self, path, destination):
    import ctypes

    LIBC = "libc.dylib"
    LIBC_FALLBACK = "/usr/lib/libSystem.dylib"
    try:
        clib = ctypes.CDLL(LIBC)
    except OSError as exc:
        logger.debug(
            f"unable to access '{LIBC}' (errno '{exc.errno}'). Falling back to '{LIBC_FALLBACK}'."
        )
        if exc.errno != errno.ENOENT:
            raise
        # NOTE: trying to bypass System Integrity Protection (SIP)
        clib = ctypes.CDLL(LIBC_FALLBACK)

    if not hasattr(clib, "clonefile"):
        return -1

    clonefile = clib.clonefile
    clonefile.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int]
    clonefile.restype = ctypes.c_int

    return clonefile(
        ctypes.c_char_p(path.encode("utf-8")),
        ctypes.c_char_p(destination.encode("utf-8")),
        ctypes.c_int(0),
    )


def humanize_bytes(
    bytes, precision=1
):  # All credit goes to: http://code.activestate.com/recipes/577081-humanized-representation-of-a-number-of-bytes/
    """Return a humanized string representation of a number of bytes.
    >>> humanize_bytes(1)
    '1 byte'
    >>> humanize_bytes(1024)
    '1.0 kB'
    >>> humanize_bytes(1024*123)
    '123.0 kB'
    >>> humanize_bytes(1024*12342)
    '12.1 MB'
    >>> humanize_bytes(1024*12342,2)
    '12.05 MB'
    >>> humanize_bytes(1024*1234,2)
    '1.21 MB'
    >>> humanize_bytes(1024*1234*1111,2)
    '1.31 GB'
    >>> humanize_bytes(1024*1234*1111,1)
    '1.3 GB'
    """
    abbrevs = (
        (1 << 50, "PB"),
        (1 << 40, "TB"),
        (1 << 30, "GB"),
        (1 << 20, "MB"),
        (1 << 10, "kB"),
        (1, "bytes"),
    )
    if bytes == 1:
        return "1 byte"
    for factor, suffix in abbrevs:
        if bytes >= factor:
            break
    return "%.*f %s" % (precision, bytes / factor, suffix)


def add_status_formatter(status, torrent_path, message):
    status_specs = {
        "seeded": ["blue", "Seeded"],
        "exists": ["yellow", "Exists"],
        "missing_files": ["red", "Missing"],
        "failed": ["magenta", "Failed"],
        "added": ["green", "Added"],
    }
    status_spec = status_specs[status]

    status_msg = f"[{click.style(status_spec[1], fg=status_spec[0])}]"
    click.echo(f" {status_msg:18s} {torrent_path.name!r} {message}")



import logging
import os
import sqlite3
from collections import namedtuple
from pathlib import Path

from .utils import decode_str, normalize_filename

logger = logging.getLogger(__name__)

COMMIT_COUNT = 10000

SeededFile = namedtuple(
    "SeededFile", ["name", "path", "download_path", "infohash", "client", "size"]
)

InsertTorrentFile = namedtuple(
    "InsertTorrentFile", ["infohash", "name", "download_path", "paths"]
)


class SearchedFile(
    namedtuple(
        "SearchedFile", ["name", "path", "size", "normalized_name", "unsplitable_root"]
    )
):
    def to_full_path(self):
        return self.path / self.name


class Database:
    _insert_counter = 0

    def __init__(self, path, utf8_compat_mode=False):
        self.db = sqlite3.connect(path)
        self.utf8_compat_mode = utf8_compat_mode
        self.create_tables()

    def create_tables(self):
        c = self.db.cursor()
        c.execute(
            """CREATE TABLE IF NOT EXISTS files (
            name varchar NOT NULL,
            path varchar NOT NULL,
            size integer NOT NULL,
            normalized_name varchar NOT NULL,
            unsplitable_root varchar,
            UNIQUE(name, path)
        )"""
        )
        c.execute(
            """CREATE INDEX IF NOT EXISTS idx_normalized_name ON files(normalized_name)"""
        )
        c.execute("""CREATE INDEX IF NOT EXISTS idx_size ON files(size)""")
        c.execute(
            """CREATE TABLE IF NOT EXISTS client_torrents (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name varchar NOT NULL,
            download_path varchar NOT NULL,
            infohash varchar NOT NULL,
            client varchar NOT NULL,
            UNIQUE(infohash, client)
        )"""
        )
        c.execute(
            """CREATE TABLE IF NOT EXISTS client_torrentfiles (
            torrent_id integer NOT NULL,
            path varchar NOT NULL,
            size integer NOT NULL,
            UNIQUE(path, torrent_id)
        )"""
        )
        self.db.commit()

    def commit(self):
        self.db.commit()
        self._insert_counter = 0

    def check_commit(self):
        self._insert_counter += 1
        if self._insert_counter >= COMMIT_COUNT:
            self.commit()

    def insert_file_path(self, path, size):
        decoded_path = decode_str(path, try_fix=self.utf8_compat_mode)
        if decoded_path is None:
            return
        name_path, name = os.path.split(decoded_path)
        normalized_name = normalize_filename(name)
        logger.debug(
            f"Inserting name: {name!r} name_path: {name_path!r} size: {size} normalized_name: {normalized_name!r}"
        )
        c = self.db.cursor()
        c.execute(
            "INSERT OR IGNORE INTO files (name, path, size, normalized_name) VALUES (?, ?, ?, ?)",
            (name, name_path, size, normalized_name),
        )
        self.check_commit()

    def mark_unsplitable_root(self, path):
        decoded_path = decode_str(path, try_fix=self.utf8_compat_mode)
        if decoded_path is None:
            return
        c = self.db.cursor()
        c.execute(
            "UPDATE files SET unsplitable_root = ? WHERE path = ? OR path LIKE ?",
            (decoded_path, decoded_path, f"{decoded_path}{os.sep}%"),
        )
        self.check_commit()

    def truncate_files(self):
        c = self.db.cursor()
        c.execute("DELETE FROM files")
        self.db.commit()

    def search_file(
        self,
        filename=None,
        size=None,
        path=None,
        normalized_filename=False,
        path_postfix=None,
        is_unsplitable=None,
        unsplitable_root=None,
    ):
        assert (
            filename is not None
            or size is not None
            or path is not None
            or normalized_filename is not None
        ), "must specify at least one argument"
        assert (
            unsplitable_root is None or is_unsplitable is None
        ), "must specify only unsplitable_root or is_unsplitable, not both"
        c = self.db.cursor()
        query, args = [], []
        if normalized_filename:
            query.append("normalized_name = ?")
            args.append(normalize_filename(normalized_filename))

        if filename:
            query.append("name = ?")
            args.append(filename)

        if size is not None:
            query.append("size = ?")
            args.append(size)

        if path is not None:
            query.append("path = ?")
            args.append(str(path))

        if path_postfix:
            path_postfix = str(path_postfix).lstrip(os.sep)
            if path_postfix != ".":
                query.append("path LIKE ?")
                args.append(f"%{os.sep}{path_postfix}")

        if is_unsplitable is not None:
            if is_unsplitable:
                query.append("unsplitable_root IS NOT NULL")
            else:
                query.append("unsplitable_root IS NULL")

        if unsplitable_root is not None:
            query.append("unsplitable_root = ?")
            args.append(str(unsplitable_root))

        query = (
            "SELECT name, path, size, normalized_name, unsplitable_root FROM files WHERE "
            + " AND ".join(query)
        )
        logger.debug(f"Doing query: {query!r} with args: {args!r}")
        return [
            SearchedFile(name, Path(path), size, normalized_name, unsplitable_root)
            for (name, path, size, normalized_name, unsplitable_root) in c.execute(
                query, args
            ).fetchall()
        ]

    def get_torrent_file_info(self, client, infohash):
        c = self.db.cursor()
        torrents = c.execute(
            "SELECT name, download_path FROM client_torrents WHERE client = ? AND infohash = ?",
            (
                client,
                infohash,
            ),
        ).fetchall()
        if not torrents:
            return None, None
        name, download_path = torrents[0]
        return name, Path(download_path)

    def insert_torrent_files_paths(self, client, insert_torrent_files):
        c = self.db.cursor()

        self.remove_torrent_files(
            client, [itf.infohash for itf in insert_torrent_files]
        )
        self.commit()

        c.executemany(
            "INSERT OR IGNORE INTO client_torrents (name, download_path, infohash, client) VALUES (?, ?, ?, ?)",
            [
                (
                    itf.name,
                    decode_str(itf.download_path, try_fix=self.utf8_compat_mode),
                    itf.infohash,
                    client,
                )
                for itf in insert_torrent_files
            ],
        )
        self.commit()

        infohash_id_mapping = dict(
            c.execute(
                f"SELECT infohash, id FROM client_torrents WHERE client = ? AND infohash IN ({','.join(['?'] * len(insert_torrent_files))})",
                (client, *[itf.infohash for itf in insert_torrent_files]),
            ).fetchall()
        )

        for itf in insert_torrent_files:
            insert_args = []
            for path, size in itf.paths:
                path = decode_str(path, try_fix=self.utf8_compat_mode)
                if path is None:
                    continue

                insert_args.append((infohash_id_mapping[itf.infohash], path, size))

            c.executemany(
                "INSERT OR IGNORE INTO client_torrentfiles (torrent_id, path, size) VALUES (?, ?, ?)",
                insert_args,
            )
        self.commit()

    def truncate_torrent_files(self, client=None):
        c = self.db.cursor()
        if client:
            c.execute(
                "DELETE FROM client_torrentfiles WHERE torrent_id IN (SELECT id FROM client_torrents WHERE client = ?)",
                (client,),
            )
            c.execute("DELETE FROM client_torrents WHERE client = ?", (client,))
        else:
            c.execute("DELETE FROM client_torrentfiles")
            c.execute("DELETE FROM client_torrents")
        self.db.commit()

    def remove_torrent_files(self, client, infohashes):
        c = self.db.cursor()
        for (id_,) in c.execute(
            f"SELECT id FROM client_torrents WHERE client = ? AND infohash IN ({','.join(['?'] * len(infohashes))})",
            (client, *infohashes),
        ):
            c.execute("DELETE FROM client_torrents WHERE id = ?", (id_,))
            c.execute("DELETE FROM client_torrentfiles WHERE torrent_id = ?", (id_,))
        self.db.commit()

    def remove_non_existing_infohashes(self, client, infohashes):
        c = self.db.cursor()
        self.remove_torrent_files(
            client,
            [
                infohash
                for infohash, in c.execute(
                    f"SELECT infohash FROM client_torrents WHERE client = ? AND infohash NOT IN ({','.join(['?'] * len(infohashes))})",
                    (client, *infohashes),
                ).fetchall()
            ],
        )

    def get_seeded_paths(self, paths):
        c = self.db.cursor()
        c.execute(
            f"""SELECT name, download_path, infohash, client, path, size FROM client_torrentfiles
                      LEFT JOIN client_torrents ON client_torrents.id = client_torrentfiles.torrent_id
                      WHERE path IN ({','.join(['?'] * len(paths))})""",
            [decode_str(p, try_fix=self.utf8_compat_mode) for p in paths],
        )

        return [
            SeededFile(name, Path(path), download_path, infohash, client, size)
            for (name, download_path, infohash, client, path, size) in c.fetchall()
        ]



import hashlib
import logging
import os
import re
import shlex
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import click
import toml
from libtc import (
    BTFailure,
    FailedToExecuteException,
    bdecode,
    bencode,
    parse_clients_from_toml_dict,
)
from libtc.utils import get_tracker_domain

from .__version__ import __version__
from .db import Database
from .exceptions import FailedToCreateLinkException
from .indexer import Indexer
from .matcher import Matcher
from .rw_cache import ReadWriteFileCache
from .utils import (
    FailedToParseTorrentException,
    PathRewriter,
    add_status_formatter,
    create_link_path,
    humanize_bytes,
    parse_torrent,
)

DEFAULT_CONFIG_FILE = """[autotorrent]
database_path = "./autotorrent.db"
link_type = "soft"
always_verify_hash = [ ]
paths = [ ]
same_paths = [ ]
add_limit_size = 128_000_000
add_limit_percent = 5
cache_touched_files = false
rw_file_cache_ttl = 86400
fast_resume = false
ignore_file_patterns = [ ]
ignore_directory_patterns = [ ]
"""

BASE_CONFIG_FILE = """[autotorrent]
database_path = "./autotorrent.db"
link_type = "soft"
always_verify_hash = [
    "*.nfo",
    "*.sfv",
    "*.diz",
]
paths = [ ]
same_paths = [ ]
add_limit_size = 128_000_000
add_limit_percent = 5
store_path = "/mnt/store_path/{client}/{torrent_name}"
skip_store_metadata = false
cache_touched_files = false
# rw_file_cache_chown = "1000:1000"
rw_file_cache_ttl = 86400
rw_file_cache_path = "/mnt/store_path/cache"
# WARNING: setting fast_resume to true can cause errors and problems.
fast_resume = false
ignore_file_patterns = [ ]
ignore_directory_patterns = [ ]

[clients]

"""

logger = logging.getLogger(__name__)


def parse_config_file(path, utf8_compat_mode=False):
    base_config = toml.loads(DEFAULT_CONFIG_FILE)
    config = toml.load(path)
    parsed_config = base_config["autotorrent"]
    parsed_config.update(config["autotorrent"])

    clients = parsed_config["clients"] = parse_clients_from_toml_dict(config)

    database_path = path.parent / Path(parsed_config["database_path"])
    parsed_config["db"] = db = Database(
        database_path, utf8_compat_mode=utf8_compat_mode
    )
    parsed_config["indexer"] = indexer = Indexer(
        db,
        ignore_file_patterns=parsed_config["ignore_file_patterns"],
        ignore_directory_patterns=parsed_config["ignore_directory_patterns"],
    )
    parsed_config["rewriter"] = rewriter = PathRewriter(parsed_config["same_paths"])
    parsed_config["matcher"] = matcher = Matcher(rewriter, db)

    rw_file_cache_chown = parsed_config.get("rw_file_cache_chown")

    if parsed_config.get("cache_touched_files"):
        parsed_config["rw_cache"] = rw_cache = ReadWriteFileCache(
            parsed_config["rw_file_cache_path"],
            parsed_config["rw_file_cache_ttl"],
            rw_file_cache_chown,
        )
    else:
        parsed_config["rw_cache"] = None

    parsed_config["fast_resume"] = parsed_config["fast_resume"]
    parsed_config["always_verify_hash"] = parsed_config["always_verify_hash"]
    parsed_config["paths"] = parsed_config["paths"]

    return parsed_config


def validate_config_path(ctx, param, value):
    if value is not None:  # check given path first
        config_path = Path(value)
        if not config_path.is_file():
            raise click.BadParameter(f"File {value!r} does not exist or is not a file.")
        return config_path

    # check the environment variables
    config_path = os.environ.get("AUTOTORRENT_CONFIG", "")
    if config_path:
        config_path = Path(config_path)
        if config_path.is_dir():
            config_path = config_path / "config.toml"
        if config_path.is_file():
            return config_path

    # path guess order
    # * ./config.toml
    # * ~/.config/autotorrent/config.toml (and windows equivalent)
    config_path = Path("config.toml")
    if config_path.is_file():
        return config_path

    config_parent_path = Path(click.get_app_dir("autotorrent"))
    if not config_parent_path.exists():
        config_parent_path.mkdir(parents=True)

    config_path = config_parent_path / Path("config.toml")
    if not config_path.exists():
        click.echo(
            f"Config file does not exist, creating an empty config file at path: {config_path!s}"
        )
        click.echo("Remember to modify it to actually do anything useful")

        config_path.write_text(BASE_CONFIG_FILE)

    return config_path


@click.group()
@click.option(
    "-c",
    "--config",
    help="Path to config file",
    callback=validate_config_path,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option("-v", "--verbose", help="Verbose logging", flag_value=True, default=False)
@click.option(
    "-u",
    "--utf8-compat-mode",
    help="Try work around utf-8 errors, not recommended",
    flag_value=True,
    default=False,
)
@click.version_option(__version__)
@click.pass_context
def cli(ctx, config, verbose, utf8_compat_mode):
    if verbose:
        logging.basicConfig(
            level=logging.DEBUG, format="%(levelname)s:%(name)s:%(lineno)d:%(message)s"
        )
    # logger.debug(f"Using config file path: {config}")
    ctx.ensure_object(dict)
    ctx.obj.update(parse_config_file(config, utf8_compat_mode=utf8_compat_mode))


@cli.command(help="See what is seeded for a given path.")
@click.option(
    "-s",
    "--summary",
    help="End the listing with a summary",
    flag_value=True,
    default=False,
)
@click.option("-d", "--depth", type=int, default=0)
@click.argument("path", nargs=-1, type=click.Path(exists=True))
@click.pass_context
def ls(ctx, summary, depth, path):
    if path:
        paths = [Path(p) for p in path]
    else:
        paths = Path(".").iterdir()

    matcher = ctx.obj["matcher"]

    stats = {
        "count": 0,
        "total_seed_size": 0,
        "total_size": 0,
    }

    def scan_paths(paths):
        for path in paths:
            p = Path(os.path.abspath(path))
            map_result = matcher.map_path_to_clients(p)
            percent = (
                map_result.total_size
                and int((map_result.seeded_size / map_result.total_size) * 100)
                or 0
            )
            if (
                map_result.total_size == map_result.seeded_size
                and map_result.total_size > 0
            ):
                color = "green"
            elif map_result.seeded_size:
                color = "yellow"
                if percent == 0:
                    percent = 1
                if percent == 100:
                    percent = 99
            else:
                color = "red"

            stats["count"] += 1
            stats["total_size"] += map_result.total_size
            stats["total_seed_size"] += map_result.seeded_size

            click.echo(
                f"[{click.style((str(percent) + '%').rjust(4), fg=color)}] {os.fsencode(path).decode(errors='replace')}"
            )

    def dive_paths(paths, depth):
        if depth <= 0:
            scan_paths(paths)
        else:
            file_paths = []
            dir_paths = []
            for path in paths:
                if path.is_dir():
                    dir_paths.append(path)
                elif path.is_file():
                    file_paths.append(path)

            scan_paths(file_paths)
            for path in dir_paths:
                dive_paths(path.iterdir(), depth - 1)

    dive_paths(paths, depth)

    if summary:
        click.echo(f"Number of paths: {stats['count']}")
        click.echo(f"Total size {humanize_bytes(stats['total_size'])}")
        click.echo(f"Total seed size: {humanize_bytes(stats['total_seed_size'])}")
        click.echo(
            f"Total unseeded size: {humanize_bytes(stats['total_size'] - stats['total_seed_size'])}"
        )


@cli.command(help="Find unseeded paths.")
@click.option(
    "-e",
    "--escape-paths",
    help="Escape the output paths",
    flag_value=True,
    default=False,
)
@click.argument("path", nargs=-1, type=click.Path(exists=True))
@click.pass_context
def find_unseeded(ctx, escape_paths, path):
    if path:
        paths = [Path(p) for p in path]
    else:
        paths = Path(".").iterdir()

    matcher = ctx.obj["matcher"]

    for path in paths:
        p = Path(os.path.abspath(path))
        map_result = matcher.map_path_to_clients(p)
        path_seeds = {}
        for f, mapped_file in map_result.files.items():
            if f.is_symlink():
                continue
            ff = f
            is_seeded = len(mapped_file.clients) > 0
            while p in ff.parents or p == ff:
                if not is_seeded and ff in path_seeds:
                    break
                if path_seeds.get(ff):
                    break
                path_seeds[ff] = is_seeded
                ff = ff.parent
        only_unseeded_paths = set(
            [f for (f, is_seeded) in path_seeds.items() if not is_seeded]
        )
        base_unseeded_paths = []
        for f in only_unseeded_paths:
            if f.parent in only_unseeded_paths:
                continue
            base_unseeded_paths.append(f)
        for unseeded_path in base_unseeded_paths:
            unseeded_path = os.path.abspath(unseeded_path)
            if escape_paths:
                unseeded_path = shlex.quote(unseeded_path)
            click.echo(unseeded_path)


@cli.command(help="Checks if the config file exists and is loadable.")
@click.pass_context
def check_config(ctx):
    click.echo("We made it this far without a crash so the config must be loadable.")


@cli.command(
    help="Remove all torrents seeding data from a path. Does not delete the actual data."
)
@click.option("-l", "--client", help="Remove from a specific client", type=str)
@click.argument("path", nargs=-1, type=click.Path(exists=True), required=True)
@click.pass_context
def rm(ctx, client, path):
    matcher = ctx.obj["matcher"]
    clients = ctx.obj["clients"]
    clients = {
        name: c["client"]
        for (name, c) in clients.items()
        if not client or name == client
    }

    if not clients:
        click.echo("No clients found")
        quit(1)

    infohashes_to_remove = {}
    for orig_p in path:
        p = Path(os.path.abspath(orig_p))
        map_result = matcher.map_path_to_clients(p)
        for mapped_file in map_result.files.values():
            for client_name, infohash in mapped_file.clients:
                if client_name not in clients:
                    continue
                infohashes_to_remove.setdefault(client_name, set()).add(infohash)

    if not infohashes_to_remove:
        click.echo("Nothing found to remove")
        quit()

    for client_name, infohashes in infohashes_to_remove.items():
        click.echo(
            f"Removing {len(infohashes)} torrent{len(infohashes) != 1 and 's' or ''} from {client_name}"
        )
        client = clients[client_name]
        for infohash in infohashes:
            client.remove(infohash)


def validate_store_path_variable(ctx, param, value):
    result = []
    for kw in value:
        k, _, v = kw.partition("=")
        if not v:
            raise click.BadParameter("format must be 'key=value'")
        result.append((k, v))
    return result


@cli.command(help="Add new torrents to a client.")
@click.argument("client", type=str)
@click.option(
    "-e",
    "--exact",
    help='Exact matching mode. Can also be considered a "reseed" mode. Disables all other modes.',
    flag_value=True,
    default=False,
)
@click.option(
    "-s",
    "--hash-probe",
    help="Probe matched files for full pieces to ensure the data matches.",
    flag_value=True,
    default=False,
)
@click.option(
    "-a",
    "--hash-size",
    help="Hash size matching mode, checks for files with same size but different filenames.",
    flag_value=True,
    default=False,
)
@click.option(
    "--print-summary",
    help="Print a summary of all actions when done.",
    flag_value=True,
    default=False,
)
@click.option("--chown", help="Chown the data folder when creating links.", type=str)
@click.option(
    "--dry-run",
    help="Do not actually create links and add the torrents, just check what would happen.",
    flag_value=True,
    default=False,
)
@click.option(
    "--move-torrent-on-add",
    help="Move the torrent to this path after it has been added successfully to the client.",
    type=click.Path(),
)
@click.option(
    "--stopped",
    help="Add the torrent in stopped state.",
    flag_value=True,
    default=False,
)
@click.option(
    "-t",
    "--store-path-template",
    help="Pass a custom template instead of using the one defined in the config file.",
)
@click.option(
    "-v",
    "--store-path-variable",
    help="Variable used for store path using a key=value syntax.",
    callback=validate_store_path_variable,
    multiple=True,
)
@click.argument("torrent", nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.pass_context  # TODO: allow feedback while running
def add(
    ctx,
    client,
    exact,
    hash_probe,
    hash_size,
    torrent,
    print_summary,
    chown,
    dry_run,
    move_torrent_on_add,
    stopped,
    store_path_template,
    store_path_variable,
):
    torrent_paths = torrent
    client_name = client
    db = ctx.obj["db"]
    matcher = ctx.obj["matcher"]
    rw_cache = ctx.obj["rw_cache"]

    clients = ctx.obj["clients"]
    client = clients.get(client_name)
    if not client:
        raise click.BadParameter(f"Unknown client: {client_name}")
    client = client["client"]
    existing_torrents = {t.infohash: t for t in client.list()}
    store_path_variables = dict(store_path_variable)
    store_path = store_path_template or ctx.obj["store_path"]
    skip_store_metadata = ctx.obj.get("skip_store_metadata", False)

    if not client:
        click.echo(f"Client {client_name} not found found")
        quit(1)

    click.echo(
        f"Matching {len(torrent_paths)} torrent{len(torrent_paths) != 1 and 's' or ''}"
    )

    if not exact and not re.findall(r"\{[^\}]+\}", store_path):
        click.echo(
            f"Store path does not contain any variables and therefore will be the same for each torrent."
        )
        quit(1)

    stats = {"seeded": 0, "added": 0, "exists": 0, "failed": 0, "missing_files": 0}
    for torrent_path in torrent_paths:
        torrent_store_path_variables = dict(store_path_variables)
        torrent_path = Path(torrent_path)
        try:
            torrent_data = bdecode(torrent_path.read_bytes())
            torrent = parse_torrent(torrent_data, utf8_compat_mode=db.utf8_compat_mode)
        except (BTFailure, FailedToParseTorrentException):
            logger.exception("Failed to parse torrent file")
            add_status_formatter("failed", torrent_path, "failed to parse torrent file")
            stats["failed"] += 1
            continue
        if torrent.trackers:
            torrent_store_path_variables["tracker_domain"] = re.sub(
                r"[\\/]", "_", get_tracker_domain(torrent.trackers[0])
            )
        if b"source" in torrent_data[b"info"]:
            torrent_store_path_variables["torrent_source"] = re.sub(
                r"[\\/]", "_", torrent_data[b"info"][b"source"].decode()
            )
        if torrent.has_file_patterns(ctx.obj["ignore_file_patterns"]):
            add_status_formatter(
                "failed",
                torrent_path,
                "file contains ignored patterns and can therefore never be matched",
            )
            stats["failed"] += 1
            continue
        infohash = hashlib.sha1(bencode(torrent_data[b"info"])).hexdigest()
        if infohash in existing_torrents:
            add_status_formatter("seeded", torrent_path, "is already seeded")
            stats["seeded"] += 1
            continue

        found_bad_hash = False
        missing_size = None
        torrent_root_path = None
        if exact:
            torrent_root_path = matcher.match_files_exact(torrent_data)
            if torrent_root_path:
                hash_verify_result, hash_touch_result = torrent.verify_hash(
                    ctx.obj["always_verify_hash"],
                    {tf.path: torrent_root_path / tf.path for tf in torrent.filelist},
                )
                if any(
                    tf
                    for (tf, tf_result) in hash_verify_result.items()
                    if tf_result != "hash-success"
                ):
                    torrent_root_path = None
                    found_bad_hash = True
        else:
            match_result = matcher.match_files_dynamic(
                torrent_data,  # TODO: double parsed here
                match_hash_size=hash_size,
                add_limit_size=ctx.obj["add_limit_size"],
                add_limit_percent=ctx.obj["add_limit_percent"],
                hash_probe=hash_probe,
            )
            missing_size = match_result.missing_size
            if match_result.success:
                hash_verify_result, hash_touch_result = torrent.verify_hash(
                    ctx.obj["always_verify_hash"], match_result.matched_files
                )
                failed_torrent_files = {
                    tf.path: tf
                    for (tf, tf_result) in hash_verify_result.items()
                    if tf_result == "hash-failed"
                }
                missing_size += sum(tf.size for tf in failed_torrent_files.values())
                max_missing_size = min(
                    ctx.obj["add_limit_size"],
                    (ctx.obj["add_limit_percent"] * torrent.size) // 100,
                )
                if max_missing_size > missing_size:
                    if ctx.obj.get("cache_touched_files"):
                        touched_torrent_paths = set(match_result.touched_files) | {
                            tf.path
                            for (tf, tf_result) in hash_touch_result.items()
                            if (
                                tf_result == "touch-success"
                                or tf_result == "touch-failed"
                                and tf.path not in failed_torrent_files
                            )
                        }
                    else:
                        touched_torrent_paths = set()

                    link_file_mapping = {}
                    for (
                        torrent_data_path,
                        actual_path,
                    ) in match_result.matched_files.items():
                        if not actual_path:
                            continue
                        if torrent_data_path in failed_torrent_files:
                            action = "copy"
                        elif torrent_data_path in touched_torrent_paths:
                            action = "cache_link"
                        else:
                            action = "link"

                        link_file_mapping[torrent_data_path] = (action, actual_path)

                    try:
                        create_link_result = create_link_path(
                            store_path,
                            link_file_mapping,
                            client_name,
                            torrent_path,
                            torrent_store_path_variables,
                            ctx.obj["link_type"],
                            rw_cache=rw_cache,
                            chown_str=chown,
                            dry_run=dry_run,
                            skip_store_metadata=skip_store_metadata,
                        )  # TODO: feedback that things take time when caching
                        if dry_run:
                            torrent_root_path = "/tmp/autotorrent_dry_run"
                        else:
                            torrent_root_path = create_link_result.data_path
                    except FailedToCreateLinkException as e:
                        logger.debug(f"Failed to create path: {e}")
                        stats["exists"] += 1
                        add_status_formatter(
                            "exists",
                            torrent_path,
                            "the link folder already exist but this torrent is not seeded by the client",
                        )
                        continue
                    except NotADirectoryError as e:
                        logger.debug(f"Failed to create path: {e}")
                        stats["failed"] += 1
                        add_status_formatter(
                            "failed",
                            torrent_path,
                            "store path is not a folder",
                        )
                        continue
                    except PermissionError as e:
                        logger.debug(f"Failed to create path: {e}")
                        stats["failed"] += 1
                        add_status_formatter(
                            "failed",
                            torrent_path,
                            "permissions on the store path restrict the ability to create links",
                        )
                        continue
                else:
                    found_bad_hash = True

        if torrent_root_path:
            if dry_run:
                add_status_formatter("added", torrent_path, "added")
                stats["added"] += 1
            else:
                try:
                    client.add(
                        torrent_data,
                        torrent_root_path,
                        fast_resume=ctx.obj["fast_resume"],
                        stopped=stopped,
                    )
                except FailedToExecuteException as e:
                    logger.debug(f"Failed to add torrent: {e!r}")
                    add_status_formatter(
                        "failed", torrent_path, "failed to send torrent to client"
                    )
                    stats["failed"] += 1
                else:
                    add_status_formatter("added", torrent_path, "added")
                    stats["added"] += 1
                    if move_torrent_on_add:
                        move_torrent_on_add = Path(move_torrent_on_add)
                        move_torrent_on_add.mkdir(exist_ok=True, parents=True)
                        torrent_path.rename(move_torrent_on_add / torrent_path.name)

        else:
            percentage_info = None
            if missing_size is not None:
                percent = (
                    torrent.size and int((1 - (missing_size / torrent.size)) * 100) or 0
                )
                if missing_size < torrent.size:
                    color = "yellow"
                    if percent == 0:
                        percent = 1
                    if percent == 100:
                        percent = 99
                else:
                    color = "red"
                percentage_info = (
                    f"with {click.style((str(percent) + '%').rjust(3), fg=color)} found"
                )
            add_status_formatter(
                "missing_files",
                torrent_path,
                f"is missing data{percentage_info and ' ' + percentage_info or ''}"
                + (found_bad_hash and " due to bad file hashes" or ""),
            )
            stats["missing_files"] += 1
            continue

    if print_summary:
        click.echo("")
        click.echo("Summary:")
        click.echo(f" Added:          {stats['added']}")
        click.echo(f" Missing files:  {stats['missing_files']}")
        click.echo(f" Failed:         {stats['failed']}")
        click.echo(f" Folder exists:  {stats['exists']}")
        click.echo(f" Already seeded: {stats['seeded']}")
        click.echo(f" Total:          {sum(stats.values())}")


@cli.command(help="Cleanup RW cache for expired items.")
@click.pass_context
def cleanup_cache(ctx):
    rw_cache = ctx.obj["rw_cache"]
    if not rw_cache:
        click.echo("No RW cached configured")
        return

    removed_paths = rw_cache.cleanup_cache()
    click.echo(
        f"Done cleaning up cache, removed {len(removed_paths)} path{len(removed_paths) != 1 and 's' or ''}"
    )


@cli.command(help="Scan your local paths files.")
@click.option(
    "-p",
    "--path",
    help="Partial scan a given path, does not remove removed files from the database.",
    type=click.Path(exists=True),
)
@click.pass_context
def scan(ctx, path):
    indexer = ctx.obj["indexer"]
    if path:
        click.echo(f"Scanning single path {path}")
        indexer.scan_paths([path], full_scan=False)
    else:
        click.echo("Doing full scan")
        indexer.scan_paths(ctx.obj["paths"], full_scan=True)
    click.echo("Done scanning")


@cli.command(help="Scan your clients for files.")
@click.option("-l", "--client", help="Scan a specific client", type=str)
@click.option(
    "-f",
    "--full",
    help="Clear old data and do a full scan",
    flag_value=True,
    default=False,
)
@click.option(
    "-a",
    "--fast",
    help="Run a fast scan, does not detect moved torrents. Overwritten by full",
    flag_value=True,
    default=False,
)
@click.pass_context
def scan_clients(ctx, client, full, fast):
    indexer = ctx.obj["indexer"]
    clients = ctx.obj["clients"]
    clients = {
        name: c["client"]
        for (name, c) in clients.items()
        if not client or name == client
    }

    if not clients:
        click.echo("No clients found")
        quit(1)

    click.echo("Scanning clients")

    indexer.scan_clients(clients, full_scan=full, fast_scan=fast)


@cli.command(help="Test your connection to your clients.")
@click.option("-l", "--client", help="Check a specific client", type=str)
@click.pass_context
def test_connection(ctx, client):
    clients = ctx.obj["clients"]
    if client:
        clients = [c for (name, c) in clients.items() if name == client]
    else:
        clients = clients.values()

    if not clients:
        click.echo("No clients found")
        quit(1)

    with ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for client in clients:
            futures.append((client, executor.submit(client["client"].test_connection)))

        for client, future in futures:
            if future.result():
                click.echo(f"{click.style('OK ', fg='green')} {client['display_name']}")
            else:
                click.echo(f"{click.style('BAD', fg='red')} {client['display_name']}")


# @cli.command(help="Remove unseeded folders from store paths")
# @click.argument("path", nargs=-1, type=click.Path(exists=True, file_okay=False, dir_okay=True))
# @click.option(
#     "-s",
#     "--skip-scan-clients",
#     help="Do not scan clients for seeded paths before finding unseeded paths.",
#     flag_value=True,
#     default=False,
# )
# @click.pass_context
# def cleanup_store_path(ctx, path, skip_scan_clients):
#     pass

# @cli.command(help="Reseed store paths")
# @click.argument("client", type=str)
# @click.argument("path", nargs=-1, type=click.Path(exists=True, file_okay=False, dir_okay=True))
# @click.option(
#     "-s",
#     "--skip-scan-clients",
#     help="Do not scan clients for seeded paths before finding unseeded paths.",
#     flag_value=True,
#     default=False,
# )
# @click.pass_context
# def reseed_store_path(ctx, path):
#     pass

# @cli.command(help="Build a bundle from torrents that can be distributed to find torrents to seed locally.")
# @click.pass_context
# def build_cross_seed_bundle(ctx):
#     pass


# @cli.command(help="Match a cross seed bundle with your scanned data.")
# @click.pass_context
# def compare_cross_seed_bundle(ctx):
#     pass

if __name__ == "__main__":
    cli()
